# TODO: remove
# from admyral.action import action
# from pydantic import BaseModel

# @action
# async def dummy_action1() -> str:
#     print("dummy action 1")
#     return "dummy action 1"

# @action
# async def dummy_action2(s: str) -> dict:
#     return {"a": s}

# @action
# async def dummy_action3() -> dict:
#     return {"a": "hello world"}

# @action
# async def dummy_action4() -> list[dict]:
#     return [
#         {"a": "hello world"},
#         {"b": "hello world"}
#     ]

# class DummyAction1Response(BaseModel):
#     x: str

# @action
# async def dummy_action5(s: str) -> DummyAction1Response:
#     return DummyAction1Response(x="Helloooooo")

# @action
# async def dummy_action6(x: int, s: str) -> dict:
#     return {"a": x, "b": s}

# @action
# def sync_dummy_action() -> str:
#     print("sync dummy action")
#     return "sync dummy action"

# @action
# async def dummy_action7(x: int, y: bool) -> str:
#     return x + 1

# @action
# def dummy_action8(s: str, i: int) -> str:
#     return s + i



from temporalio.client import Client
from temporalio.worker import Worker
from concurrent.futures import ThreadPoolExecutor

from admyral.workers.workflow_executor import WorkflowExecutor
from admyral.workers.python_executor import (
    execute_python_action,
    python_action_worker_setup,
)
from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.action_registry import ActionRegistry
from admyral.db.admyral_store import AdmyralStore
from admyral.logger import get_logger
from admyral.workers.action_executor import action_executor
from admyral.workers.workflow_run_initializer import init_workflow_run
from admyral.secret.secrets_manager import secrets_manager_factory
from admyral.workers.if_condition_executor import execute_if_condition
from admyral.utils.future_executor import capture_main_event_loop

logger = get_logger(__name__)


async def _setup():
    await python_action_worker_setup()

    admyral_store = await AdmyralStore.create_store(skip_setup=True)
    secrets_manager = secrets_manager_factory(admyral_store)
    SharedWorkerState.init(admyral_store, secrets_manager)

    capture_main_event_loop()


async def run_worker(
    worker_name: str,
    target_host: str,
    task_queue: str = "workflow-queue",
    thread_pool_size: int = 100,
    worker_debug_mode: bool = False,
) -> None:
    logger.info(f"Setting up worker {worker_name}...")
    await _setup()
    logger.info(f"Worker {worker_name} setup complete.")

    # we wrap the actions with anohter layer which automically persists the result
    activities = (
        [
            action_executor(action.action_type, action.func)
            for action in ActionRegistry.get_actions()
        ]
        + [action_executor("execute_python_action", execute_python_action)]
        + [action_executor("if_condition", execute_if_condition)]
        + [init_workflow_run]
    )

    logger.info(f"Starting worker {worker_name}...")
    client = await Client.connect(target_host)
    worker = Worker(
        client=client,
        task_queue=task_queue,
        workflows=[WorkflowExecutor],
        activities=activities,
        activity_executor=ThreadPoolExecutor(thread_pool_size),
        debug_mode=worker_debug_mode,
    )
    await worker.run()



import asyncio
import shutil
import aiofiles.os
import os


async def copyfile(src: str, dst: str):
    """
    Unfortunately, there is not true async implementation so far.
    Falling back to execute on the thread pool.
    """
    await asyncio.to_thread(shutil.copyfile, src, dst)


async def makedirs(directory: str):
    await aiofiles.os.makedirs(directory, exist_ok=True)


async def path_exists(path: str) -> bool:
    return await aiofiles.os.path.exists(path)


async def getcwd() -> str:
    return await aiofiles.os.getcwd()


async def dirname(file: str) -> str:
    return await asyncio.to_thread(os.path.dirname, file)



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_slack_client(api_key: str) -> Client:
    return Client(
        base_url="https://api.slack.com/api",
        headers={
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Send Slack Message",
    display_namespace="Slack",
    description="This method posts a message to a public channel, private channel, "
    "or direct message (DM, or IM) conversation. Required scope: chat:write",
    secrets_placeholders=["SLACK_SECRET"],
)
def send_slack_message(
    channel_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Channel ID",
            description="Channel, private group, or user to send a message to. "
            "For channels, you can use the name (e.g. #my-channel) or the ID. "
            "For private channels and groups, use the ID. For DMs to users, "
            "use the user ID.",
        ),
    ],
    text: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Text",
            description="The purpose of this field changes depends on whether "
            "the blocks field is used. If blocks is used, this is used as a "
            "fallback string to display in notifications. If blocks is not used, "
            "this is the main body text of the message. It can be formatted as "
            "plain text, or with mrkdwn.",
        ),
    ] = None,
    blocks: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Blocks", description="An array of layout blocks."
        ),
    ] = None,
) -> JsonValue:
    secret = ctx.get().secrets.get("SLACK_SECRET")
    api_key = secret["api_key"]

    body = {"channel": channel_id, "text": text, "blocks": blocks}

    with get_slack_client(api_key) as client:
        response = client.post("/chat.postMessage", json=body)
        response.raise_for_status()
        response_body = response.json()
        if not response_body.get("ok"):
            raise RuntimeError(
                f"Failed to send message in Slack. Error: {response_body.get("error")}"
            )
        return response_body


def _get_slack_user_id_by_email(client: Client, email: str, api_key: str) -> str:
    response = client.get("/users.lookupByEmail", params={"email": email})
    response.raise_for_status()
    response_body = response.json()
    if not response_body.get("ok"):
        raise RuntimeError("Failed to find user by email")
    return response_body["user"]["id"]


@action(
    display_name="Lookup User by Email",
    display_namespace="Slack",
    description="This method returns a user's identity given their email address.",
    secrets_placeholders=["SLACK_SECRET"],
)
def lookup_slack_user_by_email(
    email: Annotated[
        str,
        ArgumentMetadata(
            display_name="Email",
            description="An email address belonging to a user in the Slack workspace.",
        ),
    ],
) -> str:
    secret = ctx.get().secrets.get("SLACK_SECRET")
    api_key = secret["api_key"]

    with get_slack_client(api_key) as client:
        return _get_slack_user_id_by_email(client, email, api_key)


@action(
    display_name="Send Slack Message to User by Email",
    display_namespace="Slack",
    description="This method sends a message to a user identified by their email address.",
    secrets_placeholders=["SLACK_SECRET"],
)
def send_slack_message_to_user_by_email(
    email: Annotated[
        str,
        ArgumentMetadata(
            display_name="Email",
            description="The email address of the user to send a message to.",
        ),
    ],
    text: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Text",
            description="The purpose of this field changes depends on whether "
            "the blocks field is used. If blocks is used, this is used as a "
            "fallback string to display in notifications. If blocks is not used, "
            "this is the main body text of the message. It can be formatted as "
            "plain text, or with mrkdwn.",
        ),
    ] = None,
    blocks: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Blocks", description="An array of layout blocks."
        ),
    ] = None,
) -> JsonValue:
    # https://api.slack.com/methods/users.lookupByEmail

    secret = ctx.get().secrets.get("SLACK_SECRET")
    api_key = secret["api_key"]

    with get_slack_client(api_key) as client:
        user_id = _get_slack_user_id_by_email(client, email, api_key)

        response = client.post(
            "/chat.postMessage",
            json={"channel": user_id, "text": text, "blocks": blocks},
        )
        response.raise_for_status()
        response_body = response.json()
        if not response_body.get("ok"):
            raise RuntimeError(
                f"Failed to send message to user by email in Slack . Message: {response_body.get("error")}"
            )
        return response_body


@action(
    display_name="Batched Send Slack Message to User by Email",
    display_namespace="Slack",
    description="This method sends a batch of messages to users identified by their email address.",
    secrets_placeholders=["SLACK_SECRET"],
)
def batched_send_slack_message_to_user_by_email(
    messages: Annotated[
        list[tuple[str, str | None, JsonValue]],
        ArgumentMetadata(
            display_name="Messages",
            description="A list of messages to send to users.",
        ),
    ],
) -> JsonValue:
    for email, text, blocks in messages:
        try:
            send_slack_message_to_user_by_email(email=email, text=text, blocks=blocks)
        except Exception as e:
            print(f"Failed to send message to {email}: {e}")
            raise e



from enum import Enum
from pydantic import BaseModel
from typing import Optional, Any


class ScheduleType(str, Enum):
    CRON = "cron"
    INTERVAL_SECONDS = "interval_seconds"
    INTERVAL_MINUTES = "interval_minutes"
    INTERVAL_HOURS = "interval_hours"
    INTERVAL_DAYS = "interval_days"


class Schedule(BaseModel):
    schedule_type: ScheduleType
    value: str | int
    default_args: Optional[dict[str, Any]] = None



from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from contextlib import asynccontextmanager

from admyral.server.deps import setup_dependencies
from admyral.server.background_tasks import start_background_tasks
from admyral.config.config import API_V1_STR
from admyral.server.endpoints import (
    action_router,
    workflow_router,
    webhook_router,
    secret_router,
    editor_router,
    workflow_run_router,
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    # On Startup
    await setup_dependencies()
    start_background_tasks()
    yield
    # On Shutdown
    # ...


app = FastAPI(title="Admyral", lifespan=lifespan)


app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


app.include_router(action_router, prefix=f"{API_V1_STR}/actions")
app.include_router(workflow_router, prefix=f"{API_V1_STR}/workflows")
app.include_router(secret_router, prefix=f"{API_V1_STR}/secrets")
app.include_router(workflow_run_router, prefix=f"{API_V1_STR}/runs")
app.include_router(editor_router, prefix=f"{API_V1_STR}/editor")
app.include_router(webhook_router, prefix="/webhooks")


@app.get("/health")
async def health():
    return {"status": "ok"}



import sys
import os
import atexit
import signal
import psutil
from typing import Callable, Optional

from admyral.logger import get_logger

logger = get_logger(__name__)


def _verify_platform() -> None:
    if sys.platform == "win32":
        raise ValueError(
            "Daemons are only supported for UNIX-based operating systems and not for Windows."
        )


def _daemon_cleanup(pid_file: str, child_term_timeout: float = 5) -> None:
    """
    Cleanup function to be called when the daemon process exits.

    Args:
        pid_file: The path to the PID file.
    """
    # we first try to terminate all child processes
    proc_children = psutil.Process(os.getpid()).children(recursive=True)

    # we first try to terminate all child processes by sending them a SIGTERM signal
    for child in proc_children:
        child.terminate()
    _, alive_children = psutil.wait_procs(proc_children, timeout=child_term_timeout)

    # the still children that survived SIGTERM will be killed with SIGINT
    for child in alive_children:
        child.kill()
    psutil.wait_procs(alive_children, timeout=child_term_timeout)

    # finally, we cleanup the PID file
    if os.path.exists(pid_file):
        os.remove(pid_file)


def _get_pid_if_daemon_is_running(pid_file: str) -> Optional[int]:
    try:
        with open(pid_file, "r") as pf:
            pid = int(pf.read().strip())
    except IOError | FileNotFoundError | ValueError:
        return None

    if not pid or not psutil.pid_exists(pid):
        return None

    return pid


def launch_as_daemon(
    func: Callable[[], None],
    pid_file: str,
    log_file: Optional[str] = None,
    working_dir: Optional[str] = "/",
) -> None:
    """
    Launch a function as a daemon process using UNIX double-fork method.

    Why double fork?
    - the first fork creates a background process
    - the second fork makes sure that the daemon does not have a controlling terminal, runs independently of any terminal,
        and cannot be inadvertently attached to a terminal in the future

    Args:
        func: The function to run as a daemon.
        log_file: The path to the log file.
        pid_file: The path to the PID file.
    """
    _verify_platform()

    # get absolute paths since we change the working directory
    # we also need to make sure that the parent directories of the PID file and log file exist
    pid_file = os.path.abspath(pid_file)
    os.makedirs(os.path.dirname(pid_file), exist_ok=True)

    if log_file:
        log_file = os.path.abspath(log_file)
        os.makedirs(os.path.dirname(log_file), exist_ok=True)

    # check whether a daemon process is already running
    if os.path.exists(pid_file):
        if pid := _get_pid_if_daemon_is_running(pid_file):
            raise RuntimeError(
                f"Daemon is already running with PID {pid}. Please kill the daemon process first before trying to launch a new daemon."
            )
        # no process is running, so we can clean up the PID file
        os.remove(pid_file)

    # first, we create a child process
    try:
        pid = os.fork()
        if pid > 0:
            # exit the process which called os.fork()
            # Note: we return and do not exit here because the parent should be able to continue.
            return
    except OSError as err:
        logger.error(f"fork #1 failed: {err}")
        sys.exit(1)

    # decouple from parent environment
    os.chdir(working_dir)
    # Next, create a new session which disassociates the process from the controlling terminal.
    # Make the process the leader of a new session and a new process group.
    os.setsid()
    os.umask(0)

    # do second fork so that the new process is no longer a session leader.
    # only session leaders can acquire a controlling terminal. this makes sure that
    # the new child process can never accidentally reacquire a controlling terminal.
    try:
        pid = os.fork()
        if pid > 0:
            # exit from second parent
            sys.exit(0)
    except OSError as err:
        sys.stderr.write("fork #2 failed: {0}\n".format(err))
        sys.exit(1)

    # redirect standard file descriptors

    # flush pending output
    sys.stdout.flush()
    sys.stderr.flush()

    # open the /dev/null file (any data written to it is discarded, also it does not provide any data for reading)
    # and redirect the standard file descriptors (stdin, stdout, stderr) to it

    devnull_fd = os.open(os.devnull, os.O_RDWR)

    # if a log file is provided, we redirect stdout and stderr to the log file
    log_file_fd = (
        os.open(log_file, os.O_CREAT | os.O_RDWR | os.O_APPEND)
        if log_file
        else devnull_fd
    )

    os.dup2(devnull_fd, sys.stdin.fileno())
    os.dup2(log_file_fd, sys.stdout.fileno())
    os.dup2(log_file_fd, sys.stderr.fileno())

    # write the PID of the daemon process to the pidfile
    pid = str(os.getpid())
    with open(pid_file, "w+") as f:
        f.write(pid + "\n")

    # register the pidfile cleanup function to be called when the process exits
    atexit.register(lambda: _daemon_cleanup(pid_file))
    signal.signal(signal.SIGTERM, lambda _signum, _frame: _daemon_cleanup(pid_file))
    signal.signal(signal.SIGINT, lambda _signum, _frame: _daemon_cleanup(pid_file))

    # finally, we can now run the function as a daemon
    func()

    sys.exit(0)


def stop_daemon(pid_file: str) -> None:
    """
    Stop a running daemon process.

    Args:
        pid_file: The path to the PID file.
    """
    _verify_platform()

    if not os.path.exists(pid_file):
        logger.warning("PID file does not exist. Daemon is not running.")
        return

    pid = _get_pid_if_daemon_is_running(pid_file)
    if pid is None:
        logger.warning("Daemon is not running.")
        return

    if psutil.pid_exists(pid):
        proc = psutil.Process(pid)
        proc.terminate()
        logger.info(f"Daemon process with PID {pid} terminated.")
    else:
        logger.warning(f"Daemon process with PID {pid} does not exist.")



from typing import Annotated, Literal
from httpx import Client
import time
import random

from admyral.typings import JsonValue
from admyral.action import action, ArgumentMetadata
from admyral.context import ctx


def get_opsgenie_client(instance: str, api_key: str) -> Client:
    base_api_url = (
        "https://api.eu.opsgenie.com"
        if instance and instance.lower() == "eu"
        else "https://api.opsgenie.com"
    )
    return Client(
        base_url=base_api_url,
        headers={
            "Authorization": f"GenieKey {api_key}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Create OpsGenie Alert",
    display_namespace="OpsGenie",
    description="Create an OpsGenie alert",
    secrets_placeholders=["OPSGENIE_SECRET"],
)
def create_opsgenie_alert(
    message: Annotated[
        str,
        ArgumentMetadata(
            display_name="Message",
            description="The message of the alert. Limited to 130 characters.",
        ),
    ],
    alias: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Alias",
            description="Client-defined identifier of the alert, that is also the key element of Alert De-Duplication.",
        ),
    ] = None,
    description: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Description",
            description="Description field of the alert that is generally used to provide a detailed information about the alert.",
        ),
    ] = None,
    responders: Annotated[
        list[JsonValue] | None,
        ArgumentMetadata(
            display_name="Responders",
            description="A JSON array of teams, users, escalations, and schedules that the alert will be routed to send notifications.",
        ),
    ] = None,
    visible_to: Annotated[
        list[JsonValue] | None,
        ArgumentMetadata(
            display_name="Visible To",
            description='A JSON array of teams and users that the alert will be visible to without sending any notification. Note that the alert will be visible to the teams that are specified within "Responders" field by default, so there is no need to respecify them within the "Visible To" field',
        ),
    ] = None,
    actions: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Actions",
            description="A list of custom actions that will be available for the alert.",
        ),
    ] = None,
    tags: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Tags",
            description="A list of labels attached to the alert.",
        ),
    ] = None,
    details: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Details",
            description="Map of key-value pairs to use as custom properties of the alert.",
        ),
    ] = None,
    entity: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Entity",
            description="Entity field of the alert that is generally used to specify which domain the alert is related to.",
        ),
    ] = None,
    source: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Source",
            description="Source field of the alert. Default value is IP address of the incoming request.",
        ),
    ] = None,
    priority: Annotated[
        Literal["P1", "P2", "P3", "P4", "P5"] | None,
        ArgumentMetadata(
            display_name="Priority",
            description="Priority level of the alert. Default value is P3.",
        ),
    ] = None,
    user: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="User",
            description="Display name of the request owner.",
        ),
    ] = None,
    note: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Note",
            description="Additional note that will be added to the alert.",
        ),
    ] = None,
) -> JsonValue:
    # https://docs.opsgenie.com/docs/alert-api#section-create-alert
    opsgenie_secret = ctx.get().secrets.get("OPSGENIE_SECRET")
    api_key = opsgenie_secret["api_key"]
    instance = opsgenie_secret.get("instance")

    body = {
        "message": message,
    }
    if alias:
        body["alias"] = alias
    if description:
        body["description"] = description
    if responders:
        body["responders"] = responders
    if visible_to:
        body["visibleTo"] = visible_to
    if actions:
        body["actions"] = actions
    if tags:
        body["tags"] = tags
    if details:
        body["details"] = details
    if entity:
        body["entity"] = entity
    if source:
        body["source"] = source
    if priority:
        body["priority"] = priority
    if user:
        body["user"] = user
    if note:
        body["note"] = note

    with get_opsgenie_client(instance, api_key) as client:
        response = client.post(
            "/v2/alerts",
            json=body,
        )
        response.raise_for_status()

        response_json = response.json()
        if "requestId" not in response_json:
            raise RuntimeError("Failed to create OpsGenie alert - missing requestId.")

        return _wait_for_opsgenie_request_completion(
            client,
            response_json["requestId"],
        )


def _wait_for_opsgenie_request_completion(
    client: Client,
    request_id: str,
    max_retries: int = 10,
    cap: int = 120,
) -> JsonValue:
    for retries in range(max_retries):
        response = client.get(
            f"/v2/alerts/requests/{request_id}",
        )
        if response.status_code == 200:
            return response.json()

        # exponential backoff with full jitter
        # https://aws.amazon.com/de/blogs/architecture/exponential-backoff-and-jitter/
        if retries + 1 < max_retries:
            base_timout_ms = 5
            temp_ms = random.randrange(base_timout_ms * 2 ** (retries + 1))
            sleep_time_sec = min(cap, temp_ms / 1000)
            time.sleep(sleep_time_sec)

    raise RuntimeError(f"Failed to get OpsGenie alert: {response.text}")



from pydantic import BaseModel, Field, ConfigDict
from enum import Enum
from abc import abstractmethod, ABC
from typing import Any, Literal, Annotated

from admyral.typings import JsonValue


class IVisitor(ABC):
    @abstractmethod
    def visit(self) -> Any: ...


"""
Condition Grammer:

Constant := JsonValue
BinaryOperator := "==" | "!=" | ">" | "<" | ">=" | "<=" | "in"
UnaryOperator := "not" | "is None" | "is not None"

UnaryExpression := UnaryOperator ConditionExpression
BinaryExpression := ConditionExpression BinaryOperator ConditionExpression
AndExpression := ConditionExpression "and" ConditionExpression
OrExpression := ConditionExpression "or" ConditionExpression

ConditionExpression := "(" ConditionExpression ")" | Constant | UnaryExpression | BinaryExpression | AndExpression | OrExpression

"""


class BinaryOperator(str, Enum):
    EQUALS = "EQUALS"
    NOT_EQUALS = "NOT_EQUALS"
    GREATER_THAN = "GREATER_THAN"
    LESS_THAN = "LESS_THAN"
    GREATER_THAN_OR_EQUAL = "GREATER_THAN_OR_EQUAL"
    LESS_THAN_OR_EQUAL = "LESS_THAN_OR_EQUAL"
    IN = "IN"


class UnaryOperator(str, Enum):
    NOT = "NOT"
    IS_NONE = "IS_NONE"
    IS_NOT_NONE = "IS_NOT_NONE"


class ConditionExpression(ABC, BaseModel):
    type: str
    """ discriminator field """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def accept(self, visitor: IVisitor) -> Any:
        return None


class ConstantConditionExpression(ConditionExpression):
    type: Literal["constant"] = "constant"
    value: JsonValue

    def accept(self, visitor: IVisitor) -> Any:
        return visitor.visit(self)


class UnaryConditionExpression(ConditionExpression):
    type: Literal["unary"] = "unary"
    op: UnaryOperator
    expr: "Condition"

    def accept(self, visitor: IVisitor) -> Any:
        return visitor.visit(self)


class BinaryConditionExpression(ConditionExpression):
    type: Literal["binary"] = "binary"
    lhs: "Condition"
    op: BinaryOperator
    rhs: "Condition"

    def accept(self, visitor: IVisitor) -> Any:
        return visitor.visit(self)


class AndConditionExpression(ConditionExpression):
    type: Literal["and"] = "and"
    and_expr: list["Condition"]

    def accept(self, visitor: IVisitor) -> Any:
        return visitor.visit(self)


class OrConditionExpression(ConditionExpression):
    type: Literal["or"] = "or"
    or_expr: list["Condition"]

    def accept(self, visitor: IVisitor) -> Any:
        return visitor.visit(self)


type Condition = Annotated[
    ConstantConditionExpression
    | UnaryConditionExpression
    | BinaryConditionExpression
    | AndConditionExpression
    | OrConditionExpression,
    Field(..., discriminator="type"),
]


def condition_validate(expr: dict[str, JsonValue]) -> Condition:
    if expr["type"] == "constant":
        return ConstantConditionExpression.model_validate(expr)
    elif expr["type"] == "unary":
        return UnaryConditionExpression.model_validate(expr)
    elif expr["type"] == "binary":
        return BinaryConditionExpression.model_validate(expr)
    elif expr["type"] == "and":
        return AndConditionExpression.model_validate(expr)
    elif expr["type"] == "or":
        return OrConditionExpression.model_validate(expr)
    else:
        raise ValueError(f"Invalid condition type: {expr['type']}")



from pydantic import BaseModel
from typing import Optional

from admyral.typings import JsonValue


class WorkflowSchedule(BaseModel):
    schedule_id: str
    workflow_id: str
    cron: Optional[str] = None
    interval_seconds: Optional[int] = None
    interval_minutes: Optional[int] = None
    interval_hours: Optional[int] = None
    interval_days: Optional[int] = None
    default_args: dict[str, JsonValue] = {}



from admyral.action import action
from admyral.typings import JsonValue


# TODO: common schema for alerts?
@action(
    secrets_placeholders=["DB_URI"],
)
def load_into_database() -> JsonValue:
    raise NotImplementedError("This action is not implemented yet")



from fastapi import APIRouter, status

from admyral.models import Secret, SecretMetadata, DeleteSecretRequest
from admyral.server.deps import get_secrets_manager
from admyral.config.config import GlobalConfig


router = APIRouter()


@router.post("/set", status_code=status.HTTP_204_NO_CONTENT)
async def set_secret(secret: Secret) -> None:
    user_id = GlobalConfig().user_id  # TODO: user_id from token
    await get_secrets_manager().set(user_id, secret)


@router.get("/list", status_code=status.HTTP_200_OK)
async def list_secrets() -> list[SecretMetadata]:
    user_id = GlobalConfig().user_id  # TODO: user_id from token
    return await get_secrets_manager().list(user_id)


@router.delete("/delete", status_code=status.HTTP_204_NO_CONTENT)
async def delete_secret(request: DeleteSecretRequest) -> None:
    user_id = GlobalConfig().user_id  # TODO: user_id from token
    await get_secrets_manager().delete(user_id, request.secret_id)



from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine
from sqlmodel import SQLModel, text
from alembic.config import Config
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import RevisionStep
from alembic.script import ScriptDirectory
from pathlib import Path
import os
from typing import Callable, Optional, Iterable, Any
from sqlalchemy.engine import Connection
from functools import partial

from admyral.utils.aio import makedirs, path_exists
from admyral.config.config import GlobalConfig, DatabaseType, SQLITE_DATABASE_NAME


# TODO: why are we filtering out the alembic_version table?
def include_object(object, name, type_, reflected, compare_to):
    if type_ == "table":
        return object.schema == "admyral" and name != "alembic_version"
    return True  # Include all other types by default


def get_admyral_dir() -> str:
    return str(Path(__file__).parent.parent.parent.parent)


class DatabaseManager:
    def __init__(self, engine: AsyncEngine, config: GlobalConfig) -> None:
        self.engine = engine
        self.config = config

        self.target_metadata = SQLModel.metadata

        # setup alembic config and context
        admyral_dir = get_admyral_dir()
        self.alembic_config = Config(os.path.join(admyral_dir, "alembic.ini"))
        self.alembic_config.set_main_option(
            "script_location", os.path.join(admyral_dir, "admyral", "db", "alembic")
        )
        self.script_directory = ScriptDirectory.from_config(self.alembic_config)
        self.context = EnvironmentContext(self.alembic_config, self.script_directory)

    def _get_postgres_setup_engine(self) -> str:
        # https://stackoverflow.com/questions/6506578/how-to-create-a-new-database-using-sqlalchemy/8977109#8977109
        db_name = self.config.database_url.split("/")[-1]
        db_url = self.config.database_url[: -len(db_name)] + "postgres"
        return create_async_engine(db_url, echo=True, future=True, pool_pre_ping=True)

    async def database_exists(self) -> bool:
        if self.config.database_type == DatabaseType.SQLITE:
            return await path_exists(
                os.path.join(self.config.storage_directory, SQLITE_DATABASE_NAME)
            )

        if self.config.database_type == DatabaseType.POSTGRES:
            engine = self._get_postgres_setup_engine()
            try:
                async with engine.connect() as conn:
                    result = await conn.execute(
                        text(
                            "select exists (select 1 from pg_database where datname = 'admyral')"
                        )
                    )
                    return result.scalar()
            except Exception:
                return False

        raise NotImplementedError(
            f"Unimplemented database type in database_exists: {self.database_type}"
        )

    async def create_database(self) -> None:
        if self.config.database_type == DatabaseType.SQLITE:
            # we only need to make sure the parent directory of the SQLite file exists
            await makedirs(self.config.storage_directory)
            return

        if self.config.database_type == DatabaseType.POSTGRES:
            engine = self._get_postgres_setup_engine()
            async with engine.connect() as conn:
                await conn.execute(text("commit"))
                await conn.execute(text("create database admyral"))
            return

        raise NotImplementedError(
            f"Unimplemented database type in create_database: {self.database_type}"
        )

    async def drop_database(self) -> None:
        # TODO:
        raise NotImplementedError("Drop database not implemented yet.")

    async def migrate_database(self) -> None:
        # TODO: if the DB is not empty
        # => a lot of things could go wrong here
        # => we should probably backup the DB before we do this
        await self._upgrade()

    async def _run_migrations(
        self,
        fn: Optional[Callable[..., Iterable[RevisionStep]]],
    ) -> None:
        def _do_run_migrations(
            fn: Optional[Callable[..., Iterable[RevisionStep]]], connection: Connection
        ) -> None:
            fn_context = {}
            if fn:
                fn_context["fn"] = fn

            self.context.configure(
                connection=connection,
                target_metadata=self.target_metadata,  # TODO: understand target_metadata
                version_table_schema=self.target_metadata.schema,  # TODO: understand version_table_schema?
                include_schemas=True,  # TODO: understand include_schemas?
                include_object=include_object,
                # compare_type=True, # TODO: is this helpful?
                # render_as_batch=True, # TODO: is this helpful?
                **fn_context,
            )

            with self.context.begin_transaction():
                self.context.run_migrations()

        fn_context = {}
        if fn:
            fn_context["fn"] = fn

        async with self.engine.connect() as conn:
            await conn.run_sync(partial(_do_run_migrations, fn))

    async def _upgrade(self, revision: str = "heads") -> None:
        def do_upgrade(rev: Any, context: Any) -> list[Any]:
            return self.script_directory._upgrade_revs(
                revision,
                rev,
            )

        await self._run_migrations(do_upgrade)

    async def _downgrade(self, revision: str) -> None:
        def do_downgrade(rev: Any, context: Any) -> list[Any]:
            return self.script_directory._downgrade_revs(
                revision,
                rev,
            )

        await self._run_migrations(do_downgrade)



from typing import Annotated, Literal
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_pagerduty_client(email: str, api_key: str) -> Client:
    return Client(
        base_url="https://api.pagerduty.com",
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Token token={api_key}",
            "From": email,
        },
    )


@action(
    display_name="Create PagerDuty Incident",
    display_namespace="PagerDuty",
    description="Create a PagerDuty incident",
    secrets_placeholders=["PAGERDUTY_SECRET"],
)
def create_pagerduty_incident(
    title: Annotated[
        str,
        ArgumentMetadata(
            display_name="Title",
            description="The title of the incident",
        ),
    ],
    service_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Service ID",
            description="The ID of the service to create the incident for",
        ),
    ],
    urgency: Annotated[
        Literal["high", "low"],
        ArgumentMetadata(
            display_name="Urgency",
            description="The urgency of the incident",
        ),
    ],
    assign_to: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Assign To",
            description="List of user IDs to assign the incident to",
        ),
    ] = None,
    description: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Description",
            description="Additional incident details",
        ),
    ] = None,
) -> JsonValue:
    # https://developer.pagerduty.com/api-reference/a7d81b0e9200f-create-an-incident
    secret = ctx.get().secrets.get("PAGERDUTY_SECRET")
    api_key = secret["api_key"]
    email = secret["email"]

    body = {
        "incident": {
            "type": "incident",
            "title": title,
            "service": {"id": service_id, "type": "service_reference"},
        }
    }

    if urgency:
        body["incident"]["urgency"] = urgency

    if assign_to:
        body["incident"]["assignments"] = [
            {"assignee": {"id": user_id, "type": "user_reference"}}
            for user_id in assign_to
        ]

    if description:
        body["incident"]["body"] = {"type": "incident_body", "details": description}

    # Note: ignoring incident key and escalation policy for now

    with get_pagerduty_client(email, api_key) as client:
        response = client.post(
            "/incidents",
            json=body,
        )
        response.raise_for_status()
        return response.json()



from typing import Annotated
from httpx import Client
import base64

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_virus_total_client(api_key: str) -> Client:
    return Client(
        base_url="https://www.virustotal.com/api/v3",
        headers={
            "x-apikey": api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze File Hash",
    display_namespace="VirusTotal",
    description="Analyze a hash using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_hash(
    hash: Annotated[
        str,
        ArgumentMetadata(
            display_name="Hash",
            description="The file hash to analyze",
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    api_key = secret["api_key"]
    with get_virus_total_client(api_key) as client:
        response = client.get(f"/files/{hash}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze Domain",
    display_namespace="VirusTotal",
    description="Analyze a domain using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_domain(
    domain: Annotated[
        str,
        ArgumentMetadata(
            display_name="Domain",
            description="The domain to analyze",
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    api_key = secret["api_key"]
    with get_virus_total_client(api_key) as client:
        response = client.get(f"/domains/{domain}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze IP Address",
    display_namespace="VirusTotal",
    description="Analyze an IP address using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_ip(
    ip_address: Annotated[
        str,
        ArgumentMetadata(
            display_name="IP Address", description="The IP address to analyze"
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    api_key = secret["api_key"]
    with get_virus_total_client(api_key) as client:
        response = client.get(f"/ip_addresses/{ip_address}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze URL",
    display_namespace="VirusTotal",
    description="Analyze a URL using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_url(
    url: Annotated[
        str,
        ArgumentMetadata(
            display_name="URL",
            description="The URL to analyze",
        ),
    ],
) -> JsonValue:
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    api_key = secret["api_key"]
    with get_virus_total_client(api_key) as client:
        url_base64 = base64.b64encode(url.encode()).decode()
        response = client.get(f"/urls/{url_base64}")
        response.raise_for_status()
        return response.json()



from typing import Annotated
import json

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue


@action(
    display_name="Deserialize JSON String",
    display_namespace="Admyral",
    description="Deserializes a JSON string.",
)
def deserialize_json_string(
    serialized_json: Annotated[
        str,
        ArgumentMetadata(
            display_name="Serialized JSON String",
            description="Deserialized JSON string.",
        ),
    ],
) -> JsonValue:
    return json.loads(serialized_json)


@action(
    display_name="Serialize JSON String",
    display_namespace="Admyral",
    description="Serializes a JSON string.",
)
def serialize_json_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return json.dumps(json_value)



from admyral.actions.integrations.ai.anthropic import anthropic_chat_completion
from admyral.actions.integrations.ai.azure_openai import azure_openai_chat_completion
from admyral.actions.integrations.ai.mistralai import mistralai_chat_completion
from admyral.actions.integrations.ai.openai import openai_chat_completion

__all__ = [
    "anthropic_chat_completion",
    "azure_openai_chat_completion",
    "mistralai_chat_completion",
    "openai_chat_completion",
]



from sqlmodel import Field, Relationship
from sqlalchemy import TEXT, ForeignKeyConstraint, JSON, INTEGER
from typing import TYPE_CHECKING, Optional

from admyral.typings import JsonValue
from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import WorkflowSchedule

if TYPE_CHECKING:
    from admyral.db.schemas.workflow_schemas import WorkflowSchema


class WorkflowScheduleSchema(BaseSchema, table=True):
    """
    Schema for Workflow Schedules
    """

    __tablename__ = "workflow_schedules"

    __table_args__ = (
        ForeignKeyConstraint(
            ["workflow_id"],
            ["workflows.workflow_id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    schedule_id: str = Field(
        sa_type=TEXT(),
        primary_key=True,
    )

    # foreign keys
    user_id: str = Field(sa_type=TEXT(), index=True)
    workflow_id: str = Field(sa_type=TEXT(), index=True)

    # other fields
    cron: Optional[str] = Field(sa_type=TEXT(), nullable=True)
    interval_seconds: Optional[int] = Field(sa_type=INTEGER(), nullable=True)
    interval_minutes: Optional[int] = Field(sa_type=INTEGER(), nullable=True)
    interval_hours: Optional[int] = Field(sa_type=INTEGER(), nullable=True)
    interval_days: Optional[int] = Field(sa_type=INTEGER(), nullable=True)
    default_args: Optional[dict[str, JsonValue]] = Field(sa_type=JSON(), nullable=True)

    # relationship parents
    workflow: "WorkflowSchema" = Relationship(back_populates="schedules")

    def to_model(self, include_resources: bool = False) -> WorkflowSchedule:
        return WorkflowSchedule.model_validate(
            {
                "schedule_id": self.schedule_id,
                "workflow_id": self.workflow_id,
                "cron": self.cron,
                "interval_seconds": self.interval_seconds,
                "interval_minutes": self.interval_minutes,
                "interval_hours": self.interval_hours,
                "interval_days": self.interval_days,
                "default_args": self.default_args,
            }
        )



from typing import TYPE_CHECKING, TypeVar, Callable, Any
from temporalio import activity
import inspect
from uuid import uuid4

from admyral.context import ExecutionContext
from admyral.utils.json import throw_if_not_allowed_return_type
from admyral.utils.future_executor import execute_future
from admyral.logger import get_logger
from admyral.context import ctx
from admyral.secret.secrets_access import Secrets, SecretsStoreAccessImpl
from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.typings import JsonValue

if TYPE_CHECKING:
    F = TypeVar("F", bound=Callable[..., Any])

logger = get_logger(__name__)


def action_executor(action_type: str, func: "F") -> "F":
    if inspect.iscoroutinefunction(func):

        @activity.defn(name=action_type)
        async def async_execute(
            ctx_dict: dict, secret_mappings: dict[str, str], args: dict[str, Any]
        ) -> tuple[str, Any]:
            logger.info(f"Executing async action: {action_type}")

            exec_ctx = ExecutionContext(**ctx_dict)
            exec_ctx.step_id = str(uuid4())
            exec_ctx.secrets = Secrets(
                SecretsStoreAccessImpl(
                    user_id=exec_ctx.user_id,
                    secret_mappings=secret_mappings,
                    secrets_manager=SharedWorkerState.get_secrets_manager(),
                )
            )

            # make available to the function as a contextvar
            ctx.set(exec_ctx)

            result = await func(**args)
            throw_if_not_allowed_return_type(result)

            await _store_action_result(exec_ctx, result, args)

            return exec_ctx.step_id, result

        return async_execute

    else:

        @activity.defn(name=action_type)
        def sync_execute(
            ctx_dict: dict, secret_mappings: dict[str, str], args: dict[str, Any]
        ) -> tuple[str, Any]:
            logger.info(f"Executing sync action: {action_type}")

            exec_ctx = ExecutionContext(**ctx_dict)
            exec_ctx.step_id = str(uuid4())
            exec_ctx.secrets = Secrets(
                SecretsStoreAccessImpl(
                    user_id=exec_ctx.user_id,
                    secret_mappings=secret_mappings,
                    secrets_manager=SharedWorkerState.get_secrets_manager(),
                )
            )

            # make available to the function as a contextvar
            ctx.set(exec_ctx)

            result = func(**args)
            throw_if_not_allowed_return_type(result)

            execute_future(_store_action_result(exec_ctx, result, args))

            return exec_ctx.step_id, result

        return sync_execute


async def _store_action_result(
    exec_ctx: ExecutionContext, result: JsonValue, args: dict[str, Any]
) -> None:
    await SharedWorkerState.get_store().store_action_result(
        exec_ctx.step_id,
        exec_ctx.run_id,
        exec_ctx.action_type,
        exec_ctx.prev_step_id,
        result,
        args,
    )



from typing import TextIO


def file_tail(file: TextIO, n_lines: int) -> tuple[int, list[str]]:
    BLOCK_SIZE = 1024
    # move the cursor to the end of the file
    TO_END_OF_FILE = 2
    INITIAL_CURSOR_OFFSET = 0
    file.seek(INITIAL_CURSOR_OFFSET, TO_END_OF_FILE)

    # get current cursor position (now: end of file)
    block_end_byte = file.tell()
    available_bytes = block_end_byte

    lines_to_go = n_lines
    block_number = -1
    blocks = []

    while lines_to_go > 0 and block_end_byte > 0:
        # try to move to the next block starting from the end of the file
        # towards the beginning of the file
        if block_end_byte - BLOCK_SIZE > 0:
            # we still have sufficient bytes to read a full block

            # set the cursor to the beginning of the block
            file.seek(block_number * BLOCK_SIZE, TO_END_OF_FILE)
            blocks.append(file.read(BLOCK_SIZE))
        else:
            # the remaining bytes are less than a full block
            # so we just set the cursor to the beginning of the file
            file.seek(0, 0)
            # read the remaining bytes
            blocks.append(file.read(block_end_byte))

        # count the lines found in the current block and updates
        # the counters
        lines_found = blocks[-1].count("\n")
        lines_to_go -= lines_found
        block_end_byte -= BLOCK_SIZE
        block_number -= 1

    all_read_text = b"".join(reversed(blocks))
    tail_lines = all_read_text.decode().splitlines()[-n_lines:]

    return available_bytes, tail_lines



"""
Setup:

1, Go to the Admin Dashboard in Okta
2. Go to Security > API > Tokens
3. Click on "Create Token"

"""

from typing import Annotated
from httpx import Client, Response

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


# TODO: OAuth2: https://developer.okta.com/docs/guides/implement-oauth-for-okta-serviceapp/main/
def get_okta_client(okta_domain: str, api_key: str) -> Client:
    return Client(
        base_url=f"https://{okta_domain}/api/v1",
        headers={
            "Authorization": f"SSWS {api_key}",
            "Accept": "application/json",
        },
    )


def _get_next_link(response: Response) -> str | None:
    next_link = [
        header.split(";")[0][1:-1]
        for header in response.headers.get_list("link")
        if header.endswith('rel="next"')
    ]
    return next_link[0] if len(next_link) > 0 else None


# TODO: OCSF schema mapping
@action(
    display_name="List Events",
    display_namespace="Okta",
    description="List events from Okta",
    secrets_placeholders=["OKTA_SECRET"],
)
def list_okta_events(
    user_id: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="User ID",
            description="The user ID to list events for.",
        ),
    ] = None,
    start_time: Annotated[
        str,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "1970-01-01T00:00:00Z",
    end_time: Annotated[
        str,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "2100-01-01T00:00:00Z",
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # Note: Bounded Request - Polling Request use case not supported
    # https://developer.okta.com/docs/reference/api/system-log/

    secret = ctx.get().secrets.get("OKTA_SECRET")
    okta_domain = secret["domain"]
    api_key = secret["api_key"]

    with get_okta_client(okta_domain, api_key) as client:
        params = {
            "limit": min(limit, 1000),
            "since": start_time,
            "until": end_time,
        }
        if user_id:
            params["filter"] = f'actor.id eq "{user_id}"'

        response = client.get("/logs", params=params)
        response.raise_for_status()

        events = response.json()

        # handle pagination
        prev_num_events = len(events)
        base_url = str(client.base_url)
        while link := _get_next_link(response):
            response = client.get(link.replace(base_url, ""))
            response.raise_for_status()
            events.extend(response.json())
            if len(events) >= limit or prev_num_events == len(events):
                break
            prev_num_events = len(events)

        return events[:limit]



from admyral.models.action import Argument, PythonAction, ActionMetadata
from admyral.models.pip_lockfile import PipLockfile
from admyral.models.schedule import Schedule, ScheduleType
from admyral.models.workflow import (
    WorkflowDAG,
    WorkflowStart,
    WorkflowScheduleTrigger,
    WorkflowWebhookTrigger,
    WorkflowTriggerBase,
    WorkflowDefaultArgument,
    WorkflowTriggerType,
    IfNode,
    ActionNode,
    NodeBase,
    Workflow,
    WorkflowPushRequest,
    WorkflowPushResponse,
    WorkflowTriggerResponse,
    TriggerStatus,
    WorkflowMetadata,
)
from admyral.models.workflow_run import (
    WorkflowRun,
    WorkflowRunStep,
    WorkflowRunMetadata,
    WorkflowRunStepMetadata,
)
from admyral.models.workflow_webhook import WorkflowWebhook
from admyral.models.workflow_schedule import WorkflowSchedule
from admyral.models.secret import (
    EncryptedSecret,
    Secret,
    SecretMetadata,
    DeleteSecretRequest,
)
from admyral.models.condition import (
    IVisitor,
    BinaryOperator,
    UnaryOperator,
    ConditionExpression,
    ConstantConditionExpression,
    UnaryConditionExpression,
    BinaryConditionExpression,
    AndConditionExpression,
    OrConditionExpression,
    Condition,
    condition_validate,
)
from admyral.models.editor import (
    ActionNamespace,
    EditorActions,
    EditorWorkflowStartNode,
    EditorWorkflowActionNode,
    EditorWorkflowIfNode,
    EditorWorkflowEdgeType,
    EditorWorkflowEdge,
    EditorWorkflowGraph,
    EditorScheduleTrigger,
    EditorWebhookTrigger,
    EditorScheduleType,
)

__all__ = [
    "PipLockfile",
    "Argument",
    "PythonAction",
    "ActionMetadata",
    "Schedule",
    "ScheduleType",
    "WorkflowDAG",
    "WorkflowStart",
    "WorkflowScheduleTrigger",
    "WorkflowWebhookTrigger",
    "WorkflowTriggerBase",
    "WorkflowDefaultArgument",
    "WorkflowTriggerType",
    "IfNode",
    "IfCondition",
    "BinaryConditionExpression",
    "BinaryOperator",
    "ActionNode",
    "NodeBase",
    "Workflow",
    "WorkflowRun",
    "WorkflowRunStep",
    "WorkflowRunMetadata",
    "WorkflowRunStepMetadata",
    "WorkflowWebhook",
    "WorkflowSchedule",
    "WorkflowPushRequest",
    "WorkflowPushResponse",
    "TriggerStatus",
    "WorkflowTriggerResponse",
    "EncryptedSecret",
    "Secret",
    "SecretMetadata",
    "DeleteSecretRequest",
    "IVisitor",
    "BinaryOperator",
    "UnaryOperator",
    "ConditionExpression",
    "ConstantConditionExpression",
    "UnaryConditionExpression",
    "BinaryConditionExpression",
    "AndConditionExpression",
    "OrConditionExpression",
    "Condition",
    "condition_validate",
    "WorkflowMetadata",
    "ActionNamespace",
    "EditorActions",
    "EditorWorkflowStartNode",
    "EditorWorkflowActionNode",
    "EditorWorkflowIfNode",
    "EditorWorkflowEdgeType",
    "EditorWorkflowEdge",
    "EditorWorkflowGraph",
    "EditorScheduleTrigger",
    "EditorWebhookTrigger",
    "EditorScheduleType",
]



from typing import Optional, Literal
from pydantic import BaseModel, Field
from enum import Enum
from collections import defaultdict
from datetime import datetime

from admyral.typings import JsonValue
from admyral.models.condition import Condition


class NodeBase(BaseModel):
    id: str
    """ Unique ID of the node used for building edges. """
    type: str
    """ The type of the node. """


class ActionNode(NodeBase):
    result_name: Optional[str] = None
    args: dict[str, JsonValue] = {}
    """ The arguments of the action. Either a reference or a JSON-serialized constant. """
    secrets_mapping: dict[str, str] = {}
    """ The mapping of secret placeholders to secret IDs. """
    children: list[str] = []

    def add_edge(self, child_node: str) -> None:
        # Note: children must be a set but we can't use a set type due to the following error
        # TypeError: Object of type set is not JSON serializable
        if child_node not in self.children:
            self.children.append(child_node)

    def __str__(self) -> str:
        return f"ActionNode(id={self.id}, type={self.type}, result_name={self.result_name}, args={self.args}, secrets_mapping={self.secrets_mapping}, children={self.children})"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ActionNode):
            return False
        return (
            self.id == value.id
            and self.type == value.type
            and self.result_name == value.result_name
            and self.args == value.args
            and self.children == value.children
        )

    @classmethod
    def build_start_node(cls) -> "ActionNode":
        return cls(id="start", type="start")


class IfNode(NodeBase):
    type: Literal["if_condition"] = "if_condition"
    condition: Condition = Field(..., discriminator="type")
    condition_str: str

    # Children must be sets
    # Why?
    # if we have act5(a, b) and a, b are emitted variables from an if-condition block
    # then we connect each leaf node twice to act5(a, b) (once for a and once for b)
    true_children: list[str] = []
    false_children: list[str] = []

    def add_true_edge(self, child_node: str) -> None:
        # Note: children must be a set but we can't use a set type due to the following error
        # TypeError: Object of type set is not JSON serializable
        if child_node not in self.true_children:
            self.true_children.append(child_node)

    def add_false_edge(self, child_node: str) -> None:
        # Note: children must be a set but we can't use a set type due to the following error
        # TypeError: Object of type set is not JSON serializable
        if child_node not in self.false_children:
            self.false_children.append(child_node)

    def __str__(self) -> str:
        return f"IfNode(id={self.id}, type={self.type}, condition={self.condition}, true_children={self.true_children}, false_children={self.false_children})"

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IfNode):
            return False
        return (
            self.id == value.id
            and self.type == value.type
            and self.condition == value.condition
            and self.true_children == value.true_children
            and self.false_children == value.false_children
        )


class WorkflowTriggerType(str, Enum):
    WEBHOOK = "webhook"
    SCHEDULE = "schedule"


class WorkflowDefaultArgument(BaseModel):
    name: str
    value: JsonValue


class WorkflowTriggerBase(BaseModel):
    type: WorkflowTriggerType
    default_args: list[WorkflowDefaultArgument]

    @property
    def default_args_dict(self) -> dict[str, JsonValue]:
        return {arg.name: arg.value for arg in self.default_args}


class WorkflowWebhookTrigger(WorkflowTriggerBase):
    type: WorkflowTriggerType = WorkflowTriggerType.WEBHOOK


class WorkflowScheduleTrigger(WorkflowTriggerBase):
    type: WorkflowTriggerType = WorkflowTriggerType.SCHEDULE
    cron: Optional[str] = None
    interval_seconds: Optional[int] = None
    interval_minutes: Optional[int] = None
    interval_hours: Optional[int] = None
    interval_days: Optional[int] = None


class WorkflowStart(BaseModel):
    triggers: list[WorkflowWebhookTrigger | WorkflowScheduleTrigger] = []
    """ The triggers for the workflow. """


class WorkflowDAG(BaseModel):
    name: str
    description: str | None = None
    start: WorkflowStart
    dag: dict[str, IfNode | ActionNode]
    version: str = "1"

    # TODO: make this a property
    def get_in_deg(self) -> dict[str, int]:
        in_deg = defaultdict(int)
        for node in self.dag.values():
            if isinstance(node, IfNode):
                for true_child in node.true_children:
                    in_deg[true_child] += 1
                for false_child in node.false_children:
                    in_deg[false_child] += 1
            else:
                assert isinstance(node, ActionNode)
                for child in node.children:
                    in_deg[child] += 1
        return in_deg


class Workflow(BaseModel):
    workflow_id: str
    workflow_name: str
    workflow_dag: WorkflowDAG
    is_active: bool


class WorkflowPushRequest(BaseModel):
    workflow_dag: WorkflowDAG
    activate: bool


class WorkflowPushResponse(BaseModel):
    webhook_id: Optional[str] = None
    webhook_secret: Optional[str] = None


class TriggerStatus(str, Enum):
    SUCCESS = "SUCCESS"
    INACTIVE = "INACTIVE"


class WorkflowTriggerResponse(BaseModel):
    status: TriggerStatus

    @classmethod
    def success(cls) -> "WorkflowTriggerResponse":
        return cls(status=TriggerStatus.SUCCESS)

    @classmethod
    def inactive(cls) -> "WorkflowTriggerResponse":
        return cls(status=TriggerStatus.INACTIVE)


class WorkflowMetadata(BaseModel):
    workflow_id: str
    workflow_name: str
    workflow_description: str | None = None
    created_at: datetime
    is_active: bool



from typing import Annotated
from httpx import Client

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue


def get_snyk_client(region: str, api_token: str) -> Client:
    return Client(
        base_url=_get_base_api_url(region),
        headers={
            "Authorization": f"token {api_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


# TODO: OCSF mapping
@action(
    display_name="List Vulnerabilities",
    display_namespace="Snyk",
    description="List Snyk vulnerabilities",
    secrets_placeholders=["SNYK_SECRET"],
)
def list_snyk_vulnerabilities(
    org_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Organization ID",
            description="The ID of the organization to list vulnerabilities for",
        ),
    ],
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time of the vulnerabilities in ISO-8601 format (inclusive)",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time of the vulnerabilities in ISO-8601 format (inclusive)",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of vulnerabilities to return.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://apidocs.snyk.io/?version=2024-06-21#get-/orgs/-org_id-/issues

    secret = ctx.get().secrets.get("SNYK_SECRET")
    api_token = secret["api_token"]
    region = secret.get("region")

    params = {}
    if start_time:
        params["created_after"] = start_time
    if end_time:
        params["created_before"] = end_time
    if limit:
        params["limit"] = limit

    with get_snyk_client(region, api_token) as client:
        response = client.get(f"/rest/orgs/{org_id}/issues", params=params)
        response.raise_for_status()

        # handle pagination
        result = response.json()
        data = result.get("data", [])

        next_url = _get_next_url(result)
        while next_url:
            next_url = next_url.replace(str(client.base_url), "")
            response = client.get(
                next_url,
            )
            response.raise_for_status()
            result = response.json()

            data.extend(result.get("data", []))
            next_url = _get_next_url(result)

        return data


def _get_next_url(result: dict) -> str | None:
    if "links" in result and "next" in result["links"]:
        next_url = result["links"]["next"]
        if not isinstance(next_url, str):
            next_url = next_url["href"]
        return next_url
    return None


def _get_base_api_url(region: str) -> str:
    match region.lower():
        case "au":
            return "https://api.au.snyk.io"
        case "eu":
            return "https://api.eu.snyk.io"
        case _:
            return "https://api.snyk.io"
